#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>
#include <math.h>

typedef enum channel_type {
	voltage_ch,
	resistor_ch,
	inamp_ch,
	photo_ch,
	wheatstone_ch
} channel_type;

typedef enum quantity_type {
	voltage,
	resistance,
	temp_kelvin,
	temp_celsius
} quantity_type;


const float ADC_RES = 0.001220703f;
const int PULLUP_RESISTANCE = 10000;
const int THERM_R_25 = 10000;
const int THERM_B = 3977;

void error_sent_and_received_not_match(unsigned char sent, unsigned char received)
{
	printf("Error: '%c' (%d) was received instead of '%c'! The program now exit...\n", received, received, sent);
	exit(EXIT_FAILURE);
}

unsigned char send_and_receive(int fd, unsigned char c)
{
	unsigned char received = 0, ret = 0;
	write(fd,&c,1); 
	while (ret == 0)
		ret = read(fd,&received,1);
	return received;
}

unsigned char send_receive_check(int fd, unsigned char c)
{
	unsigned char received = 0, ret = 0;
	write(fd,&c,1);
	while (ret == 0) {
		ret = read(fd,&received,1);
	}
	
	if(received == c)
		return received;
	else
		error_sent_and_received_not_match(c, received);
}

void set_edaq_comm(int fd)
{
	struct termios tio;
 
        memset(&tio,0,sizeof(tio));
        tio.c_iflag=0;
        tio.c_oflag=0;
        tio.c_cflag=CS8|CREAD|CLOCAL;           // 8n1, see termios.h for more information
        tio.c_lflag=0;
        tio.c_cc[VMIN]=1;
        tio.c_cc[VTIME]=5;
 
        cfsetospeed(&tio,B230400);            // 230400 baud
        cfsetispeed(&tio,B230400);            // 230400 baud
 
        tcsetattr(fd,TCSANOW,&tio);
}

int open_edaq(char* name)
{
	int fd = open(name, O_RDWR | O_NOCTTY | O_SYNC);
	//int fd = open(name, O_RDWR | O_NOCTTY | O_NONBLOCK);
	set_edaq_comm(fd);
	return fd;
}



void print_edaq_id(int fd)
{
	unsigned char c;
	c = send_receive_check(fd, '@');

	c = send_receive_check(fd, 'I');

	printf("Device identifier:\t");
	c = 'a';
	const unsigned char sent = '#';
	while(c != sent)
	{	
		c = send_and_receive(fd, sent);
		if(c != sent)
			printf("%c", c);
	}
    printf("\n");
    fflush(stdout);
}

unsigned char get_channel_source(channel_type ch)
{
	switch (ch) 
	{	
	case voltage_ch:
		return 0;
	case resistor_ch:
		return 0;
	case inamp_ch:
		return 1;
	case photo_ch:
		return 2;
	case wheatstone_ch:
		return 0;	//don't know jet
	default: 
		return 0;
	}
	
}

void set_inputs(int fd, channel_type ch_a, channel_type ch_b, channel_type ch_c)
{
	unsigned char c = 0;	
	c = send_receive_check(fd, '@');
	c = send_receive_check(fd, 'C');
	c = send_receive_check(fd, get_channel_source(ch_a));
	c = send_receive_check(fd, get_channel_source(ch_b));
	c = send_receive_check(fd, get_channel_source(ch_c));
	
	c = send_receive_check(fd, '@');
	c = send_receive_check(fd, 'E');
	if (ch_a == photo_ch || ch_b == photo_ch || ch_c == photo_ch)
		c = send_receive_check(fd, 1);
	else
		c = send_receive_check(fd, 0);
	//pullup/dn ???

	c = send_receive_check(fd, '@');
	c = send_receive_check(fd, 'P');
	c = 0;
	if (ch_a != resistor_ch)
		c |= 1;
	if (ch_b != resistor_ch)
		c |= 2;
	if (ch_c != resistor_ch)
		c |= 4;

	c = send_receive_check(fd, c);	
}

//11<=sampling_freq<=1000, 3 channels
void set_sampling_freq(int fd, unsigned int sampling_freq)
{
	unsigned char c = 0;	
	c = send_receive_check(fd, '@');
	c = send_receive_check(fd, 'f');
	c = (unsigned char) 0;
	c = send_receive_check(fd, c);
	
	sampling_freq *= 3;
	c = (unsigned char) (sampling_freq >> 8);
	c = send_receive_check(fd, c);
	c = (unsigned char) sampling_freq;
	c = send_receive_check(fd, c);
}

void start_cont_sampling(int fd)
{
	unsigned char c = 0;	
	c = send_receive_check(fd, '@');
	c = send_receive_check(fd, 'S');
}

void fill_array_cont_sampling(int fd, int* data, int count)
{
	int i, j;
	unsigned char c;
	for (i = 0; i < count; i++)
	{
		read(fd, &c, 1);
		data[i] = c * 256;
		read(fd, &c, 1);
		data[i] += c;
		for (j = 0; j < 4; j++)
			read(fd, &c, 1);
		
	}

}

void stop_cont_sampling(int fd)
{
	unsigned char c = 27; //ESC
	write(fd, &c, 1);
	tcflush(fd, TCIFLUSH);
}

void measure_edaq(int fd, int* data)
{
	unsigned char c = 0;
	int i = 0;
	c = send_receive_check(fd, '@');

	c = send_receive_check(fd, 'M');

	c = send_receive_check(fd, '#');
	
	for(i = 0; i< 3;i++)
	{
		read(fd,&c,1);
		data[i] = ((int)c)<<8;
		read(fd,&c,1);
		data[i] += c;
	}
}


float adc_code_to_voltage(int d)
{
	return d * ADC_RES;
}


float adc_code_to_resistance(int d)
{
	float u = adc_code_to_voltage(d);
	u /= 5.0f;
	return u * (PULLUP_RESISTANCE / (1.0f - u));
}

float adc_code_to_temp_kelvin(int d)
{
	float r = adc_code_to_resistance(d);
	return 1.0f / ((1.0f / 298.16f) + ((1.0f / THERM_B) * (float) log((double) (r / THERM_R_25))));
}

float adc_code_to_temp_celsius(int d)
{
	return adc_code_to_temp_kelvin(d) - 273.16f;
}

float get_measured_quantity(int d, quantity_type type)
{
	switch (type)
	{
	case voltage:
		return adc_code_to_voltage(d);
	case resistance:
		return adc_code_to_resistance(d);
	case temp_kelvin:
		return adc_code_to_temp_kelvin(d);
	case temp_celsius:
		return adc_code_to_temp_celsius(d);
	default:
		return adc_code_to_voltage(d);
	}
}






